#define PIN_RELAY   A3  // Пин, к которому подключено реле
#define PIN_BUZZER  A1  // Пин, к которому подключен буззер

// Подключение библиотек для работы с LCD по протоколу I2C
#include <Wire.h> 
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);
//lcd.write((uint8_t)0);

// Массивы пользовательских (нестандартных) символов для дисплея
uint8_t arrowUp[8] = {0x04, 0x0E, 0x1F, 0x04, 0x04, 0x04, 0x04, 0x00};    // Символ стрелки "вверх"
uint8_t arrowDown[8] = {0x04, 0x04, 0x04, 0x04, 0x1F, 0x0E, 0x04, 0x00};  // Символ стрелки "вниз"

#include <avr/pgmspace.h> // Библиотека для хранения строк во флеш-памяти микроконтроллера
// Определение текстовых строк меню во флеш-памяти микроконтроллера
const char menuStr_0[] PROGMEM = "ARDUINO ACCESS";
const char menuStr_1[] PROGMEM = "CONTROL SYSTEM";
const char menuStr_2[] PROGMEM = "Attach RFID-key";
const char menuStr_3[] PROGMEM = "Password:";
const char menuStr_4[] PROGMEM = "ACCESS ALOWED";
const char menuStr_5[] PROGMEM = "Lock after";
const char menuStr_6[] PROGMEM = "ACCESS DENIED";
const char menuStr_7[] PROGMEM = "Bad key/password";
const char menuStr_8[] PROGMEM = "EDIT RFID DATA";
const char menuStr_9[] PROGMEM = "A  B  #-ok *-esc";
const char menuStr_10[] PROGMEM = "EDIT<  > [A]-add";
const char menuStr_11[] PROGMEM = "[C]-clear  *-esc";
const char menuStr_12[] PROGMEM = "READING CARD UID";
const char menuStr_13[] PROGMEM = "Weating for card";
const char menuStr_14[] PROGMEM = "CARD IS DELETED";
const char menuStr_15[] PROGMEM = "FROM EEPROM";
const char menuStr_16[] PROGMEM = "ACCESS PASSWORD";
const char menuStr_17[] PROGMEM = "Set new access";
const char menuStr_18[] PROGMEM = "password:";
const char menuStr_19[] PROGMEM = "ADMIN PASSWORD";
const char menuStr_20[] PROGMEM = "Set new admin";
const char menuStr_21[] PROGMEM = "SET OPEN TIME";
const char menuStr_22[] PROGMEM = "Set open timer:";

// Создание таблицы строк
const char* const menuStringTable[] PROGMEM = { menuStr_0, menuStr_1, menuStr_2, menuStr_3,
                                                menuStr_4, menuStr_5, menuStr_6, menuStr_7,
                                                menuStr_8, menuStr_9, menuStr_10, menuStr_11,
                                                menuStr_12, menuStr_13, menuStr_14, menuStr_15,
                                                menuStr_16, menuStr_17, menuStr_18, menuStr_19,
                                                menuStr_20, menuStr_21, menuStr_22};

// Массив-буффер для извлечения строки и последующей работы с ней
char menuStringBuffer[16];

// Подключение библиотеки для работы с матричной клавиатурой
#include <Keypad.h>
#define ROWS  4 // Количество рядов матричной клавиатуры
#define COLS  4 // Количество колонок матричной клавиатуры

// Массив-карта матричной клавиатуры
char keys[ROWS][COLS] = {
  {'1','2','3','A'},
  {'4','5','6','B'},
  {'7','8','9','C'},
  {'*','0','#','D'}
};
byte rowPins[ROWS] = {A0, 8, 7, 6}; // Подключение к Arduino рядов матричной клавиатуры
byte colPins[COLS] = {5, 4, 3, 2};  // Подключение к Arduino колонок матричной клавиатуры
Keypad keypad = Keypad(makeKeymap(keys), rowPins, colPins, ROWS, COLS); // Экземпляр класса кл-ры

// Подключение библиотек для работы с RFID-модулем RC522
#include <SPI.h>
#include <MFRC522.h>
#define PIN_SS   10
#define PIN_RST  9
MFRC522 mfrc522(PIN_SS, PIN_RST); // Создание экземпляра класса для модуля RC522

/* ГЛОБАЛЬНЫЕ ПЕРЕМЕННЫЕ И ФЛАГИ */
bool flagClearMenuScreen = true;
uint32_t openTime = 5;  // Время, через которое закроется замок его открытия
uint8_t tempPassword[7] = {0, 0, 0, 0, 0, 0, 0}; // Массив для временного хранения введённого пароля
int passwordIndex = -1; // Индекс цифры пароля, с которой мы в данный момент работаем
char key; // Переменная для хранения кода нажатой клавиши
uint8_t cards[30][5]; // Массив для хранения идентификаторов RFID-карт
uint8_t cardsIndex = 0; // Индекс для работы с конкретной RFID-картой
uint8_t accessPassword[7] = {7, 6, 5, 4, 3, 2, 1}; // Пароль доступа по усолчанию
uint8_t adminPassword[7] = {1, 2, 3, 4, 5, 6, 7}; // Пароль администратора по умолчанию
uint8_t menuFlag = 0; // Номер экрана меню
uint8_t globalState = 0; // Глобальное состояние системы (для реализации конечного автомата)
uint32_t cardWeatingTime; // Время ожидания карты для регистрации

// Библиоте кдля работы с энергонезавичимой памятью EEPROM
#include <EEPROM.h>

/* ФУНКЦИЯ ИНИЦИАЛИЗАЦИИ ЭНЕРГОНЕЗАВИСИМОЙ ПАМЯТИ */
void initEEPROM() {
  // Если мы запускаем программу впервые (за данный контроль будет отвечать ячейка №500)
  if(EEPROM.read(500) != 84) { // Магическое число 84, как признак первичного запуска
    // Инициализируем нулями ячейки, отвечающие за признак хранения значений карт
    for(uint8_t i = 0; i < 30; i++) {
      EEPROM.write((i * 5), 0);
    }
    // Записываем пароль доступа и администратора по усмолчанию
    for(uint8_t i = 0; i < 7; i++) {
      EEPROM.write(i + 150, accessPassword[i]);
      EEPROM.write(i + 157, adminPassword[i]);
    }
    EEPROM.write(164, openTime); // Запись по умолчанию времени удержания замка в откритом состоянии
    EEPROM.write(500, 84); // Запись признака первого запуска программы
  }
  else {  // Если программа запускается не впервые, читаем данные из EEPROM
    for(uint8_t i = 0; i < 7; i++) {
      accessPassword[i] = EEPROM.read(i + 150);
      adminPassword[i] = EEPROM.read(i + 157);
    }
    openTime = EEPROM.read(164);

    for(uint8_t i = 0; i < 30; i++) {
      cards[i][0] = EEPROM.read(i * 5);
      cards[i][1] = EEPROM.read(i * 5 + 1);
      cards[i][2] = EEPROM.read(i * 5 + 2);
      cards[i][3] = EEPROM.read(i * 5 + 3);
      cards[i][4] = EEPROM.read(i * 5 + 4);
    }
  }
}

/* 
   КАРТА ПАМЯТИ EEPROM
   [порядковый номер карты][ (идентификатор пустой ячейки 0-пустая; 1-записана), (бит UID0), (бит UID1), (бит UID2), (бит UID3) ]

   Карты распологаются по следующим адресам:
   Карта №01 - 0000, 0001, 0002, 0003, 0004
   Карта №02 - 0005, 0006, 0007, 0008, 0009
   Карта №03 - 0010, 0011, 0012, 0013, 0014
   Карта №04 - 0015, 0016, 0017, 0018, 0019
   Карта №05 - 0020, 0021, 0022, 0023, 0024
   Карта №06 - 0025, 0026, 0027, 0028, 0029
   Карта №07 - 0030, 0031, 0032, 0033, 0034
   Карта №08 - 0035, 0036, 0037, 0038, 0039
   Карта №09 - 0040, 0041, 0042, 0043, 0044
   Карта №10 - 0045, 0046, 0047, 0048, 0049
   Карта №11 - 0050, 0051, 0052, 0053, 0054
   Карта №12 - 0055, 0056, 0057, 0058, 0059
   Карта №13 - 0060, 0061, 0062, 0063, 0064
   Карта №14 - 0065, 0066, 0067, 0068, 0069
   Карта №15 - 0070, 0071, 0072, 0073, 0074
   Карта №16 - 0075, 0076, 0077, 0078, 0079
   Карта №17 - 0080, 0081, 0082, 0083, 0084
   Карта №18 - 0085, 0086, 0087, 0088, 0089
   Карта №19 - 0090, 0091, 0092, 0093, 0094
   Карта №20 - 0095, 0096, 0097, 0098, 0099
   Карта №21 - 0100, 0101, 0102, 0103, 0104
   Карта №22 - 0105, 0106, 0107, 0108, 0109
   Карта №23 - 0110, 0111, 0112, 0113, 0114
   Карта №24 - 0115, 0116, 0117, 0118, 0119
   Карта №25 - 0120, 0121, 0122, 0123, 0124
   Карта №26 - 0125, 0126, 0127, 0128, 0129
   Карта №27 - 0130, 0131, 0132, 0133, 0134
   Карта №28 - 0135, 0136, 0137, 0138, 0139
   Карта №29 - 0140, 0141, 0142, 0143, 0144
   Карта №30 - 0145, 0146, 0147, 0148, 0149
 
   Прочее:
   Пароль доступа: 0150, 0151, 0152, 0153, 0154, 0155, 0156
   Пароль админа:  0157, 0158, 0159, 0160, 0161, 0162, 0163
   Время удержания замка в открытом состоянии: 0164
 */

/* ФУНКЦИЯ ЧТЕНИЯ ДАННЫХ О ЗАПИСАННЫХ КАРТАХ ИЗ EEPROM */
void readCardBaseFromEEPROM() {
  // Заполняем массив нулевыми значениями
  for(uint8_t i = 0; i < 30; i++) {
    for(uint8_t j = 0; j < 5; j++) {
      cards[i][j] = 0;
    }
  }
  // Чтение данных из EEPROM и заполнение массива
  for(int i = 0; i < 30; i++) {
    if(EEPROM.read((i * 5)) == 0) {
      cards[i][0] = 0;
      cards[i][1] = 0;
      cards[i][2] = 0;
      cards[i][3] = 0;
      cards[i][4] = 0;
    }
    else {
      cards[i][0] = EEPROM.read((i * 5));
      cards[i][1] = EEPROM.read((i * 5 + 1));
      cards[i][2] = EEPROM.read((i * 5 + 2));
      cards[i][3] = EEPROM.read((i * 5 + 3));
      cards[i][4] = EEPROM.read((i * 5 + 4));
    }
  }
}

/* ФУНКЦИЯ СКАНИРОВАНИЯ ПОДНЕСЁННОЙ К ТЕРМИНАЛУ КАРТЫ-ИЛИ КЛЮЧА */
// Возврат значений: 1-ключ подходит; 2-ключ отсутствует в базе
void compareCardWithDataBase(bool resetArray = true) {
  if(!mfrc522.PICC_IsNewCardPresent()) return 0;
  if(!mfrc522.PICC_ReadCardSerial()) return 0;
  // Сканируем поднесённую карту и сравниваем её с одой из карт, внесённых в базу
  for(int i = 0; i < 30; i++) {
    if((cards[i][1] == mfrc522.uid.uidByte[0]) && 
       (cards[i][2] == mfrc522.uid.uidByte[1]) && 
       (cards[i][3] == mfrc522.uid.uidByte[2]) && 
       (cards[i][4] == mfrc522.uid.uidByte[3])) {
      globalState = 3; return 0;
    }
    else globalState = 4;
  }
}

/* ФУНКЦИЯ ВЫВОДА СТРОКИ ИЗ ФЛЕШ-ПАМЯТИ МИКРОКОНТРОЛЛЕРА В ОПРЕДЕЛЁННЫЕ КООРДИНАТЫ ЖКИ */
void flashStringToLcd(uint8_t x, uint8_t y, uint8_t stringNumber) { 
  strcpy_P(menuStringBuffer, (char*)pgm_read_word(&(menuStringTable[stringNumber])));
  lcd.setCursor(x, y);
  lcd.print(menuStringBuffer);
}

/* ФУНКЦИЯ ПОДАЧИ ЗВУКОВОГО СИГНАЛА */
void beep(int l, uint8_t p) {
  for(int i = 0; i < l; i++) {
    digitalWrite(PIN_BUZZER, HIGH);
    delay(p);
    digitalWrite(PIN_BUZZER, LOW);
    delay(p);
  }  
}

/* ФУНКЦИЯ СБРОСА БУФФЕРНОГО ПАРОЛЯ */
void resetTempPassword() {
  for(uint8_t i = 0; i < 7; i++) tempPassword[i] = 0;
  passwordIndex = -1;
}

/* ФУНКЦИЯ ВЫВОДА МИГАЮЩЕЙ ИНФОРМАЦИИ, ГОВОРЯЩЕЙ ОБ ОШИБКЕ ДОСТУПА */
void accesDenied() {
  for(uint8_t i = 0; i < 3; i++) {
    beep(100, 1);
    flashStringToLcd(1, 0, 6); // ACCESS DENIED!
    flashStringToLcd(0, 1, 7); // Bad key/password
    delay(500);
    lcd.clear();
  }
}

/* ФУНКЦИЯ ОТКРЫТИЯ ЭЛЕКТРОЗАМКА ПРИ ПРАВИЛЬНОМ КОДЕ ДОСТУПА */
void accessAllowed() {
  static uint8_t localState = 0;
  static uint8_t counter = openTime;
  if(localState == 0) {
    beep(100, 1);
    digitalWrite(PIN_RELAY, HIGH); // Выключем реле, открываем электрозамок
    flashStringToLcd(1, 0, 4);    // ACCESS ALOWED
    flashStringToLcd(0, 1, 5);    // Lock after
    localState = 1;
  }
  else if(localState == 1) {
    lcd.setCursor(11, 1);
    lcd.print(counter);
    lcd.print(" sec");
    localState = 2;
  }
  else if(localState == 2) {
    if(millis() % 1000 == 0) {
      counter--;
      if(counter == 0) localState = 3; else localState = 1;
    }
  }
  else if(localState == 3) {
    digitalWrite(PIN_RELAY, LOW);  // Закрываем замок по истечении времени
    lcd.clear();
    counter = openTime;
    globalState = 0;
    localState = 0;
  }
}

/* ФУНКЦИЯ ВВОДА ПАРОЛЯ */
bool enterPassword() {
  if(key == '*') { // Стираем пароль
    for(uint8_t i = 0; i < 7; i++) tempPassword[i] = 0;
    passwordIndex = -1;
    lcd.setCursor(9, 1); lcd.print("       ");
  }
  else if(key == '#') {
    Serial.println("ENTER");
    if(passwordIndex == 6) return 1;
  }
  else {
    lcd.setCursor(passwordIndex + 10, 1);
    if(passwordIndex <= 6) {
      lcd.print(key);
      if(passwordIndex == 6) { lcd.setCursor(15, 1); lcd.print(key); }
      if(passwordIndex < 6) passwordIndex++;
      tempPassword[passwordIndex] = (key - 48); // Заносим очередной символ во временный массив
    }
  }
  return 0;
}

/* ФУНКЦИЯ ОТОБРАЖЕНИЯ МЕНЮ  */
void showMenu(uint8_t menuIndex) {
  if(flagClearMenuScreen) { lcd.clear(); flagClearMenuScreen = false; }
  switch(menuIndex) {
    case 0:
      flashStringToLcd(0, 0, 2);  // Attach RFID-key
      flashStringToLcd(0, 1, 3);  // Password:
    break;
    case 10:
      flashStringToLcd(0, 0, 8);  // EDIT RFID DATA
      flashStringToLcd(0, 1, 9);  // A  B  #-ok *-esc
      lcd.setCursor(1, 1); lcd.write((uint8_t)1); // Стрелка вверх
      lcd.setCursor(4, 1); lcd.write((uint8_t)2); // Стрелка вниз
    break;
      case 101:
        lcd.setCursor(0, 0);
        lcd.print("<");
        if(cardsIndex < 10) lcd.print("0");
        lcd.print(cardsIndex);
        lcd.print(">-");
        if(cards[cardsIndex][0] == 0) lcd.print("empty");
        else {
          lcd.print(cards[cardsIndex][1], HEX); lcd.print(":");
          lcd.print(cards[cardsIndex][2], HEX); lcd.print(":");
          lcd.print(cards[cardsIndex][3], HEX); lcd.print(":");
          lcd.print(cards[cardsIndex][4], HEX);
        }
        flashStringToLcd(0, 1, 9);  // A  B  #-ok *-esc
        lcd.setCursor(1, 1); lcd.write((uint8_t)1); // Стрелка вверх
        lcd.setCursor(4, 1); lcd.write((uint8_t)2); // Стрелка вниз
      break;
        case 102:
          lcd.setCursor(0, 0);
          lcd.print("EDIT<");
          if(cardsIndex < 10) lcd.print("0");
          lcd.print(cardsIndex);
          lcd.print("> [A]-add");
          lcd.setCursor(0, 1);
          lcd.print("[C]-clear  *-esc");
        break;
          case 103:
            lcd.setCursor(0, 0); lcd.print("READING CARD UID");
            lcd.setCursor(0, 1); lcd.print("Weating for card");
          break;
    case 11:
      flashStringToLcd(0, 0, 16); // ACCESS PASSWORD
      flashStringToLcd(0, 1, 9);  // A  B  #-ok *-esc
      lcd.setCursor(1, 1); lcd.write((uint8_t)1); // Стрелка вверх
      lcd.setCursor(4, 1); lcd.write((uint8_t)2); // Стрелка вниз
    break;
      case 111:
        flashStringToLcd(0, 0, 17); // Set new access
        flashStringToLcd(0, 1, 18); // password:
      break;
    case 12:
      flashStringToLcd(0, 0, 19);  // ADMIN PASSWORD
      flashStringToLcd(0, 1, 9);  // A  B  #-ok *-esc
      lcd.setCursor(1, 1); lcd.write((uint8_t)1); // Стрелка вверх
      lcd.setCursor(4, 1); lcd.write((uint8_t)2); // Стрелка вниз
    break;
      case 121:
        flashStringToLcd(0, 0, 20); // Set new admin
        flashStringToLcd(0, 1, 18); // password:
      break;
    case 13:
      flashStringToLcd(0, 0, 21);  // SET OPEN TIME
      flashStringToLcd(0, 1, 9);  // A  B  #-ok *-esc
      lcd.setCursor(1, 1); lcd.write((uint8_t)1); // Стрелка вверх
      lcd.setCursor(4, 1); lcd.write((uint8_t)2); // Стрелка вниз
    break;
      case 131:
        flashStringToLcd(0, 0, 22);  // Set open timer:
        lcd.setCursor(0, 1); lcd.print(">"); lcd.print(openTime); lcd.print("sec");
      break;
  }
}

/* ФУНКЦИЯ ПРЕДВАРИТЕЛЬНЫХ УСТАНОВОК */
void setup() {
  Serial.begin(9600);
  // Инициализация аппаратного SPI и параметров модуля RC522 
  SPI.begin(); 
  mfrc522.PCD_Init();
  mfrc522.PCD_SetAntennaGain(mfrc522.RxGain_max); // Максимальное усиление антенны модуля
  
  lcd.begin(16, 2);               // Инициализация дисплея 16х2
  lcd.createChar (1, arrowUp);    // Создание символа стрелки "вверх"
  lcd.createChar (2, arrowDown);  // Создание символа стрелки "вниз"

  pinMode(PIN_RELAY, OUTPUT); digitalWrite(PIN_RELAY, LOW); // Обестачиваем реле
  
  // Вывод приветствия на дисплее
  flashStringToLcd(1, 0, 0); flashStringToLcd(1, 1, 1);
  
  // Генерация короткого звукового сигнала
  pinMode(PIN_BUZZER, OUTPUT);
  beep(500, 2);

  lcd.clear();

  initEEPROM(); // Инициализация базы данных RFID-ключей
}

/* ОСНОВНОЙ ЦИКЛ ПРОГРАММЫ */
void loop() {
  while(1) {
    key = keypad.getKey(); // Чтение кода нажатой клавиши
    
    // Отображаем главный экран с приглашением ввода пароля или поднесения карты
    if(globalState == 0) {
      showMenu(0);
      globalState = 1;
    }
    // Обрабатываем ввод пароля или поднесения RFID-ключа
    else if(globalState == 1) {
      // Обрабатываем поднесённый к терминалу RFID-ключ
      compareCardWithDataBase();

      // Обрабатываем ввод пароля на матричной клавиатуре
      if(key) { 
        if(enterPassword()) { lcd.clear(); globalState = 2; }
      }
    }
    // Проверка на правильность ввода пароля
    else if(globalState == 2) {
      // Если введён пароль доступа к объекту
      if((tempPassword[0] == accessPassword[0]) && (tempPassword[1] == accessPassword[1]) &&
         (tempPassword[2] == accessPassword[2]) && (tempPassword[3] == accessPassword[3]) &&
         (tempPassword[4] == accessPassword[4]) && (tempPassword[5] == accessPassword[5]) &&
         (tempPassword[6] == accessPassword[6])) { lcd.clear(); globalState = 3; }
      // Если введён пароль доступа в меню администратора
      else if((tempPassword[0] == adminPassword[0]) && (tempPassword[1] == adminPassword[1]) &&
         (tempPassword[2] == adminPassword[2]) && (tempPassword[3] == adminPassword[3]) &&
         (tempPassword[4] == adminPassword[4]) && (tempPassword[5] == adminPassword[5]) &&
         (tempPassword[6] == adminPassword[6])) { lcd.clear(); globalState = 100; menuFlag = 10;}
      // Если пароль не соответствует ни одному из зарезервированных в системе
      else {
        // Уходим в состояние обработки ошибки
        globalState = 4;
      }
      resetTempPassword(); // Сбрасываем буффер временного пароля
    }
    // Открываем элетрозамок на определённое в меню время
    else if(globalState == 3) { accessAllowed(); } // Функция открытия замка
    // Вывод ошибки при наборе неправильного пароля или незарегистрированного ключа
    else if(globalState == 4) {
      accesDenied(); // Функция вывода ошибки доступа
      globalState = 0; // Возврат в первоначальне меню рабочего режима
    }
    // Главное меню администратора
    else if(globalState == 100) {
      showMenu(menuFlag);
      globalState = 101;
    }
    // Обработка клавиатуры в главном меню администратора
    else if(globalState == 101) {
      if(key) {
        if(key == 'A') { // Листаем вверх главное меню администратора
          key = 'n';
          menuFlag--;
          if(menuFlag < 10) menuFlag = 13;
          globalState = 100;
        }
        else if(key == 'B') { // Листаем вниз главное меню администратора
          key = 'n';
          menuFlag++;
          if(menuFlag > 13) menuFlag = 10;
          globalState = 100;
        }
        else if(key == '*') { // Выход в рабочий режим
          key = 'n';
          globalState = 0;
        }
        else if(key == '#') { // Нажатие ОК в одном из пунктов главного меню
          switch(menuFlag) { // Проверяем, в каком именно пункте мы нажали ОК
            case 10: globalState = 110; cardsIndex = 0; break;
            case 11: globalState = 120; break;
            case 12: globalState = 130; break;
            case 13: globalState = 140; break;
          }
        }
        lcd.clear();
      }
    }
    // Переходим в режим просмотра зарегистрированных в системе карт
    else if(globalState == 110) { showMenu(101); globalState = 111; }
    else if(globalState == 111) {
      // Обрабатываем ввод пароля на матричной клавиатуре
      if(key) {
        if(key == 'A') { // Листаем вверх список RFID-ключей
          key = 'n';
          cardsIndex--;
          if(cardsIndex == 255) cardsIndex = 29;
          globalState = 110;
        }
        else if(key == 'B') { // Листаем вниз список RFID-ключей
          key = 'n';
          cardsIndex++;
          if(cardsIndex > 29) cardsIndex = 0;
          globalState = 110;
        }
        else if(key == '*') { // Выход в рабочий режим
          cardsIndex = 0;
          globalState = 100;
        }
        else if(key == '#') {
          globalState = 112;
        }
        lcd.clear();
      }
    }
    // Переходим в подменю редактирования записи о RFID-ключе
    else if(globalState == 112) { showMenu(102); globalState = 113;}
    else if(globalState == 113) {
      // Определяем, что делать с выбранной записью
      if(key) {
        if(key == '*') { globalState = 110; }
        else if(key == 'C') {
          cards[cardsIndex][0] = 0;
          cards[cardsIndex][1] = 0;
          cards[cardsIndex][2] = 0;
          cards[cardsIndex][3] = 0;
          cards[cardsIndex][4] = 0;
          
          // Удаляем карту из EEPROM
          EEPROM.write(cardsIndex * 5, 0);
          EEPROM.write(cardsIndex * 5 + 1, 0);
          EEPROM.write(cardsIndex * 5 + 2, 0);
          EEPROM.write(cardsIndex * 5 + 3, 0);
          EEPROM.write(cardsIndex * 5 + 4, 0);
          
          lcd.clear(); lcd.setCursor(0, 0); lcd.print("RFID-key cleared");
          delay(2000);
          globalState = 110;
        }
        else if(key == 'A') {
          globalState = 114;
        }
      lcd.clear();
      }
    }
    else if(globalState == 114) {
      showMenu(103);
      cardWeatingTime = millis();
      globalState = 115;
    }
    // Ожидаем поднесения RFID-ключа к терминалу для последующей регистрации
    else if(globalState == 115) {
      if((millis() - cardWeatingTime) > 10000) { 
        beep(100, 1); 
        lcd. clear(); 
        globalState = 110;
      }
      if(!mfrc522.PICC_IsNewCardPresent()) return 0;
      if(!mfrc522.PICC_ReadCardSerial()) return 0;  
      beep(100, 1);
      cards[cardsIndex][0] = 1;
      cards[cardsIndex][1] = mfrc522.uid.uidByte[0];
      cards[cardsIndex][2] = mfrc522.uid.uidByte[1];
      cards[cardsIndex][3] = mfrc522.uid.uidByte[2];
      cards[cardsIndex][4] = mfrc522.uid.uidByte[3];
      
      // Запись идентификатора ключа в EEPROM
      EEPROM.write(cardsIndex * 5, 1);
      EEPROM.write(cardsIndex * 5 + 1, cards[cardsIndex][1]);
      EEPROM.write(cardsIndex * 5 + 2, cards[cardsIndex][2]);
      EEPROM.write(cardsIndex * 5 + 3, cards[cardsIndex][3]);
      EEPROM.write(cardsIndex * 5 + 4, cards[cardsIndex][4]);
      
      lcd.clear();
      globalState = 110;
    }
    
    // Переходим в режим смены пароля доступа
    else if(globalState == 120) { showMenu(111); globalState = 121; }
    else if(globalState == 121) {
      // Обрабатываем ввод пароля на матричной клавиатуре
      if(key) { 
        if(enterPassword()) {
          // Сохраняем новый пароль в EEPROM
          for(uint8_t i = 0; i < 7; i++) {
            accessPassword[i] = tempPassword[i];
            EEPROM.write(i + 150, accessPassword[i]);
          }
          resetTempPassword();  // Сбрасываем буффер временного пароля 
          lcd.clear();
          menuFlag = 11;        // Возвращаемся в предыдущее меню
          globalState = 100;    
        }
      }
    }
    // Переходим в режим смены пароля администратора
    else if(globalState == 130) { showMenu(121); globalState = 131; }
    else if(globalState == 131) {
      // Обрабатываем ввод пароля на матричной клавиатуре
      if(key) { 
        if(enterPassword()) {
          // Сохраняем новый пароль в EEPROM
          for(uint8_t i = 0; i < 7; i++) {
            adminPassword[i] = tempPassword[i];
            EEPROM.write(i + 157, adminPassword[i]);
          }
          resetTempPassword();  // Сбрасываем буффер временного пароля 
          lcd.clear();
          menuFlag = 12;        // Возвращаемся в предыдущее меню
          globalState = 100;    
        }
      }    
    }
    // Переходим в режим смены времени удержания замка в открытом состоянии
    else if(globalState == 140) { showMenu(131); globalState = 141; }
    else if(globalState == 141) {
      if(key) {
        if(key == '*' || key == '#' || key == 'A' || key == 'B' || key == 'C' || key == 'D') {
          key = 'n';
          lcd.clear();
          menuFlag = 13;
          globalState = 100;
          EEPROM.write(164, openTime);
        }
        else {
          openTime = (key - 48);
          lcd.setCursor(1, 1); lcd.print(key);
        }
      }
    }
    
  } // while(1)
}
// **************************************************************************************************
